@@ -1,7 +1,7 @@
 /*
  * dm-verity volume handling
  *
- * Copyright (C) 2012-2017, Red Hat, Inc. All rights reserved.
+ * Copyright (C) 2012-2018, Red Hat, Inc. All rights reserved.
  *
  * This file is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -66,19 +66,19 @@ int VERITY_read_sb(struct crypt_device *cd,
 		sizeof(struct verity_sb), device_path(device), sb_offset);
 
 	if (params->flags & CRYPT_VERITY_NO_HEADER) {
-		log_err(cd, _("Verity device %s doesn't use on-disk header.\n"),
+		log_err(cd, _("Verity device %s doesn't use on-disk header."),
 			device_path(device));
 		return -EINVAL;
 	}
 
 	if (sb_offset % 512) {
-		log_err(cd, _("Unsupported VERITY hash offset.\n"));
+		log_err(cd, _("Unsupported VERITY hash offset."));
 		return -EINVAL;
 	}
 
 	devfd = device_open(device, O_RDONLY);
 	if (devfd < 0) {
-		log_err(cd, _("Cannot open device %s.\n"), device_path(device));
+		log_err(cd, _("Cannot open device %s."), device_path(device));
 		return -EINVAL;
 	}
 
@@ -91,19 +91,19 @@ int VERITY_read_sb(struct crypt_device *cd,
 	close(devfd);
 
 	if (memcmp(sb.signature, VERITY_SIGNATURE, sizeof(sb.signature))) {
-		log_err(cd, _("Device %s is not a valid VERITY device.\n"),
+		log_err(cd, _("Device %s is not a valid VERITY device."),
 			device_path(device));
 		return -EINVAL;
 	}
 
 	sb_version = le32_to_cpu(sb.version);
 	if (sb_version != 1) {
-		log_err(cd, _("Unsupported VERITY version %d.\n"), sb_version);
+		log_err(cd, _("Unsupported VERITY version %d."), sb_version);
 		return -EINVAL;
 	}
 	params->hash_type = le32_to_cpu(sb.hash_type);
 	if (params->hash_type > VERITY_MAX_HASH_TYPE) {
-		log_err(cd, _("Unsupported VERITY hash type %d.\n"), params->hash_type);
+		log_err(cd, _("Unsupported VERITY hash type %d."), params->hash_type);
 		return -EINVAL;
 	}
 
@@ -111,7 +111,7 @@ int VERITY_read_sb(struct crypt_device *cd,
 	params->hash_block_size = le32_to_cpu(sb.hash_block_size);
 	if (VERITY_BLOCK_SIZE_OK(params->data_block_size) ||
 	    VERITY_BLOCK_SIZE_OK(params->hash_block_size)) {
-		log_err(cd, _("Unsupported VERITY block size.\n"));
+		log_err(cd, _("Unsupported VERITY block size."));
 		return -EINVAL;
 	}
 	params->data_size = le64_to_cpu(sb.data_blocks);
@@ -120,7 +120,7 @@ int VERITY_read_sb(struct crypt_device *cd,
 	if (!params->hash_name)
 		return -ENOMEM;
 	if (crypt_hash_size(params->hash_name) <= 0) {
-		log_err(cd, _("Hash algorithm %s not supported.\n"),
+		log_err(cd, _("Hash algorithm %s not supported."),
 			params->hash_name);
 		free(CONST_CAST(char*)params->hash_name);
 		params->hash_name = NULL;
@@ -129,7 +129,7 @@ int VERITY_read_sb(struct crypt_device *cd,
 
 	params->salt_size = le16_to_cpu(sb.salt_size);
 	if (params->salt_size > sizeof(sb.salt)) {
-		log_err(cd, _("VERITY header corrupted.\n"));
+		log_err(cd, _("VERITY header corrupted."));
 		free(CONST_CAST(char*)params->hash_name);
 		params->hash_name = NULL;
 		return -EINVAL;
@@ -166,20 +166,20 @@ int VERITY_write_sb(struct crypt_device *cd,
 		sizeof(struct verity_sb), device_path(device), sb_offset);
 
 	if (!uuid_string || uuid_parse(uuid_string, uuid) == -1) {
-		log_err(cd, _("Wrong VERITY UUID format provided on device %s.\n"),
+		log_err(cd, _("Wrong VERITY UUID format provided on device %s."),
 			device_path(device));
 		return -EINVAL;
 	}
 
 	if (params->flags & CRYPT_VERITY_NO_HEADER) {
-		log_err(cd, _("Verity device %s doesn't use on-disk header.\n"),
+		log_err(cd, _("Verity device %s doesn't use on-disk header."),
 			device_path(device));
 		return -EINVAL;
 	}
 
 	devfd = device_open(device, O_RDWR);
 	if (devfd < 0) {
-		log_err(cd, _("Cannot open device %s.\n"), device_path(device));
+		log_err(cd, _("Cannot open device %s."), device_path(device));
 		return -EINVAL;
 	}
 
@@ -199,7 +199,7 @@ int VERITY_write_sb(struct crypt_device *cd,
 	r = write_lseek_blockwise(devfd, device_block_size(device), device_alignment(device),
 				  (char*)&sb, hdr_size, sb_offset) < hdr_size ? -EIO : 0;
 	if (r)
-		log_err(cd, _("Error during update of verity header on device %s.\n"),
+		log_err(cd, _("Error during update of verity header on device %s."),
 			device_path(device));
 	close(devfd);
 
@@ -242,6 +242,7 @@ int VERITY_activate(struct crypt_device *cd,
 {
 	struct crypt_dm_active_device dmd;
 	uint32_t dmv_flags;
+	unsigned int fec_errors = 0;
 	int r;
 
 	log_dbg("Trying to activate VERITY device %s using hash %s.",
@@ -249,8 +250,18 @@ int VERITY_activate(struct crypt_device *cd,
 
 	if (verity_hdr->flags & CRYPT_VERITY_CHECK_HASH) {
 		log_dbg("Verification of data in userspace required.");
-		r = VERITY_verify(cd, verity_hdr,
-				  root_hash, root_hash_size);
+		r = VERITY_verify(cd, verity_hdr, root_hash, root_hash_size);
+
+		if (r == -EPERM && fec_device) {
+			log_dbg("Verification failed, trying to repair with FEC device.");
+			r = VERITY_FEC_process(cd, verity_hdr, fec_device, 1, &fec_errors);
+			if (r < 0)
+				log_err(cd, _("Errors cannot be repaired with FEC device."));
+			else if (fec_errors)
+				log_err(cd, _("Found %u repairable errors with FEC device."),
+					fec_errors);
+		}
+
 		if (r < 0)
 			return r;
 	}
@@ -291,7 +302,7 @@ int VERITY_activate(struct crypt_device *cd,
 
 	r = dm_create_device(cd, name, CRYPT_VERITY, &dmd, 0);
 	if (r < 0 && (dm_flags(DM_VERITY, &dmv_flags) || !(dmv_flags & DM_VERITY_SUPPORTED))) {
-		log_err(cd, _("Kernel doesn't support dm-verity mapping.\n"));
+		log_err(cd, _("Kernel doesn't support dm-verity mapping."));
 		return -ENOTSUP;
 	}
 	if (r < 0)
@@ -302,6 +313,6 @@ int VERITY_activate(struct crypt_device *cd,
 		return r;
 
 	if (!r)
-		log_err(cd, _("Verity device detected corruption after activation.\n"));
+		log_err(cd, _("Verity device detected corruption after activation."));
 	return 0;
 }
